import json

CREAM_LENS_ABI = json.loads('''
[
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken",
                "name": "cToken",
                "type": "address"
            },
            {
                "internalType": "address payable",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "cTokenBalances",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "balanceOf",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "borrowBalanceCurrent",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "balanceOfUnderlying",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "tokenBalance",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "tokenAllowance",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenBalances",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken[]",
                "name": "cTokens",
                "type": "address[]"
            },
            {
                "internalType": "address payable",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "cTokenBalancesAll",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "balanceOf",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "borrowBalanceCurrent",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "balanceOfUnderlying",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "tokenBalance",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "tokenAllowance",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenBalances[]",
                "name": "",
                "type": "tuple[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken",
                "name": "cToken",
                "type": "address"
            }
        ],
        "name": "cTokenMetadata",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "exchangeRateCurrent",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "supplyRatePerBlock",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "borrowRatePerBlock",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "reserveFactorMantissa",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalBorrows",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalReserves",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalSupply",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalCash",
                        "type": "uint256"
                    },
                    {
                        "internalType": "bool",
                        "name": "isListed",
                        "type": "bool"
                    },
                    {
                        "internalType": "uint256",
                        "name": "collateralFactorMantissa",
                        "type": "uint256"
                    },
                    {
                        "internalType": "address",
                        "name": "underlyingAssetAddress",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "cTokenDecimals",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "underlyingDecimals",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenMetadata",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken[]",
                "name": "cTokens",
                "type": "address[]"
            }
        ],
        "name": "cTokenMetadataAll",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "exchangeRateCurrent",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "supplyRatePerBlock",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "borrowRatePerBlock",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "reserveFactorMantissa",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalBorrows",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalReserves",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalSupply",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalCash",
                        "type": "uint256"
                    },
                    {
                        "internalType": "bool",
                        "name": "isListed",
                        "type": "bool"
                    },
                    {
                        "internalType": "uint256",
                        "name": "collateralFactorMantissa",
                        "type": "uint256"
                    },
                    {
                        "internalType": "address",
                        "name": "underlyingAssetAddress",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "cTokenDecimals",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "underlyingDecimals",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenMetadata[]",
                "name": "",
                "type": "tuple[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken",
                "name": "cToken",
                "type": "address"
            }
        ],
        "name": "cTokenUnderlyingPrice",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "underlyingPrice",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenUnderlyingPrice",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CToken[]",
                "name": "cTokens",
                "type": "address[]"
            }
        ],
        "name": "cTokenUnderlyingPriceAll",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "cToken",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "underlyingPrice",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CTokenUnderlyingPrice[]",
                "name": "",
                "type": "tuple[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract ComptrollerLensInterface",
                "name": "comptroller",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getAccountLimits",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "contract CToken[]",
                        "name": "markets",
                        "type": "address[]"
                    },
                    {
                        "internalType": "uint256",
                        "name": "liquidity",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "shortfall",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.AccountLimits",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CCakeLPInterface[]",
                "name": "cTokens",
                "type": "address[]"
            },
            {
                "internalType": "address",
                "name": "cake",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getClaimableCakeRewards",
        "outputs": [
            {
                "internalType": "uint256[]",
                "name": "",
                "type": "uint256[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CCTokenInterface[]",
                "name": "cTokens",
                "type": "address[]"
            },
            {
                "internalType": "address",
                "name": "comp",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getClaimableCompRewards",
        "outputs": [
            {
                "internalType": "uint256[]",
                "name": "",
                "type": "uint256[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract CSLPInterface[]",
                "name": "cTokens",
                "type": "address[]"
            },
            {
                "internalType": "address",
                "name": "sushi",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getClaimableSushiRewards",
        "outputs": [
            {
                "internalType": "uint256[]",
                "name": "",
                "type": "uint256[]"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [
            {
                "internalType": "contract Comp",
                "name": "comp",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getCompBalanceMetadata",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "uint256",
                        "name": "balance",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "votes",
                        "type": "uint256"
                    },
                    {
                        "internalType": "address",
                        "name": "delegate",
                        "type": "address"
                    }
                ],
                "internalType": "struct CompoundLens.CompBalanceMetadata",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            {
                "internalType": "contract Comp",
                "name": "comp",
                "type": "address"
            },
            {
                "internalType": "contract ComptrollerLensInterface",
                "name": "comptroller",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "getCompBalanceMetadataExt",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "uint256",
                        "name": "balance",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "votes",
                        "type": "uint256"
                    },
                    {
                        "internalType": "address",
                        "name": "delegate",
                        "type": "address"
                    },
                    {
                        "internalType": "uint256",
                        "name": "allocated",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CompBalanceMetadataExt",
                "name": "",
                "type": "tuple"
            }
        ],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [
            {
                "internalType": "contract Comp",
                "name": "comp",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            },
            {
                "internalType": "uint32[]",
                "name": "blockNumbers",
                "type": "uint32[]"
            }
        ],
        "name": "getCompVotes",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "uint256",
                        "name": "blockNumber",
                        "type": "uint256"
                    },
                    {
                        "internalType": "uint256",
                        "name": "votes",
                        "type": "uint256"
                    }
                ],
                "internalType": "struct CompoundLens.CompVotes[]",
                "name": "",
                "type": "tuple[]"
            }
        ],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    }
]
''')



import logging
from query_state_lib.client.client_querier import ClientQuerier
from web3 import Web3
from web3.middleware import geth_poa_middleware
from defi_services.abis.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.contract_address import WrappedNativeTokens
from defi_services.utils.batch_queries_service import add_rpc_call, decode_data_response
from defi_services.utils.graph_operations import BlockTimestampGraph, GraphOperations
from defi_services.utils.market_service import MarketService
from defi_services.utils.memory_storage import MemoryStorage

_LOGGER = logging.getLogger("StateService")


class StateService:
    def __init__(self, provider_uri: str):
        self._w3 = Web3(Web3.HTTPProvider(provider_uri))
        self.client_querier = ClientQuerier(provider_url=provider_uri)
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        timestamp_graph = BlockTimestampGraph(self._w3)
        self._market_service = MarketService()
        self._graph_operations = GraphOperations(timestamp_graph)
        self.memory_storage = MemoryStorage.get_instance()

    def get_latest_block(self):
        return self._w3.eth.block_number

    def get_block_nearest_timestamp(self, timestamp: int):
        try:
            start_block_bounds = self._graph_operations.get_bounds_for_y_coordinate(timestamp)
        except Exception as ex:
            _LOGGER.exception(ex)
            start_block_bounds = (0, 0)
        return start_block_bounds[0]

    def to_checksum(self, address: str):
        return self._w3.toChecksumAddress(address.lower())

    def balance_of(self, address: str, token: str, block_number: int = 'latest'):
        token_contract = self._w3.eth.contract(self.to_checksum(token), abi=ERC20_ABI)
        decimals = token_contract.functions.decimals().call()
        balance = token_contract.functions.balanceOf(self.to_checksum(address)).call(
            block_identifier=block_number)
        return balance / 10 ** decimals

    def total_supply(self, token_address: str, block_number: int = 'latest'):
        token_contract = self._w3.eth.contract(self.to_checksum(token_address), abi=ERC20_ABI)
        decimals = token_contract.functions.decimals().call()
        total_supply = token_contract.functions.totalSupply().call(block_identifier=block_number) / 10 ** decimals
        return total_supply

    def decimals(self, token_address: str):
        token_contract = self._w3.eth.contract(self.to_checksum(token_address), abi=ERC20_ABI)
        decimals = token_contract.functions.decimals().call()
        return decimals

    def decimals_of_token_list(self, token_addresses: list, block_number: int = 'latest'):
        list_rpc_call, list_call_id = [], []
        for token_address in token_addresses:
            add_rpc_call(
                abi=ERC20_ABI, fn_name='decimals', contract_address=self.to_checksum(token_address),
                block_number=block_number, list_rpc_call=list_rpc_call, list_call_id=list_call_id
            )
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        result = {}
        for token_address in token_addresses:
            decimals_call_id = f'decimals_{token_address}_{block_number}'.lower()
            result[token_address.lower()] = decoded_data.get(decimals_call_id)

        return result

    def total_supply_of_token_list(self, token_addresses: list, block_number: int = 'latest'):
        list_rpc_call, list_call_id = [], []
        for token_address in token_addresses:
            add_rpc_call(
                abi=ERC20_ABI, fn_name='totalSupply', contract_address=self.to_checksum(token_address),
                block_number=block_number, list_rpc_call=list_rpc_call, list_call_id=list_call_id
            )
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        result = {}
        for token_address in token_addresses:
            total_supply_call_id = f'totalSupply_{token_address}_{block_number}'.lower()
            result[token_address.lower()] = decoded_data.get(total_supply_call_id)

        return result



from defi_services.abis.erc20_abi import ERC20_ABI
from defi_services.abis.lending_pool.ctoken_abi import CTOKEN_ABI
from defi_services.abis.lending_pool.venus_comptroller_abi import VENUS_COMPTROLLER_ABI
from defi_services.abis.lending_pool.venus_lens_abi import VENUS_LENS_ABI
from defi_services.constants.contract_address import ContractAddresses, AbnormalVenusPool, WrappedNativeTokens
from defi_services.lending_pools.services.cream_state_service import CreamStateService
from defi_services.utils.batch_queries_service import add_rpc_call, decode_data_response


class VenusStateService(CreamStateService):
    def __init__(self, provider_uri):
        super().__init__(provider_uri)

    def vtoken_metadata_all(
            self, lens_address: str, lens_abi: list, token_addresses: list, block_number: int = 'latest'):
        lens_contract = self._w3.eth.contract(
            address=self.to_checksum(lens_address), abi=lens_abi
        )
        tokens = [self.to_checksum(i) for i in token_addresses]
        reserve_tokens_info = lens_contract.functions.vTokenMetadataAll(tokens).call(block_identifier=block_number)
        return reserve_tokens_info

    def vtoken_underlying_price_all(
            self, lens_address: str, lens_abi: list, token_addresses: list, bnb_price: float,
            block_number: int = 'latest'):
        lens_contract = self._w3.eth.contract(
            address=self.to_checksum(lens_address), abi=lens_abi
        )
        tokens = [self.to_checksum(i) for i in token_addresses]
        reserve_tokens_info = lens_contract.functions.vTokenUnderlyingPriceAll(tokens).call(
            block_identifier=block_number)
        for price_token in reserve_tokens_info:
            reserve_tokens_info[price_token[0].lower()] = price_token[1] * bnb_price
        return reserve_tokens_info

    def get_reserves_info(
            self,
            comptroller_address: str,
            comptroller_abi: list,
            lens_address: str,
            lens_abi: list,
            block_number: int = 'latest'):
        ctokens = self.get_all_markets(comptroller_address, comptroller_abi, block_number)
        metadata = self.vtoken_metadata_all(lens_address, lens_abi, ctokens, block_number)
        reserves_info = {}
        for data in metadata:
            underlying = data[11].lower()
            ctoken = data[0].lower()
            lt = data[10] / 10 ** 18
            reserves_info[underlying] = {
                "cToken": ctoken,
                "liquidationThreshold": lt
            }

        return reserves_info

    def get_token_speed(
            self, token_addresses: list, comptroller: str,
            comptroller_abi: list = VENUS_COMPTROLLER_ABI, block_number: int = 'latest'):
        list_rpc_call = []
        list_call_id = []
        for token_address in token_addresses:
            add_rpc_call(
                abi=comptroller, fn_name='venusSpeeds', contract_address=comptroller_abi,
                block_number=block_number, fn_paras=token_address, list_rpc_call=list_rpc_call,
                list_call_id=list_call_id
            )
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        result = {}
        for token_address in token_addresses:
            get_asset_data_call_id = f'venusSpeeds_{comptroller}_{token_address}_{block_number}'.lower()
            result[token_address.lower()] = decoded_data.get(get_asset_data_call_id)

        return result

    @staticmethod
    def _encode_apy_lending_pool_function_call(
            lens_address: str,
            lens_abi: list,
            comptroller_address: str,
            comptroller_abi: list,
            token_addresses: list,
            block_number: int = "latest"
    ):
        list_call_id, list_rpc_call = [], []
        for token_address in token_addresses:
            add_rpc_call(abi=lens_abi, contract_address=lens_address, fn_paras=token_address, block_number=block_number,
                         list_call_id=list_call_id, list_rpc_call=list_rpc_call, fn_name="vTokenMetadata")
            add_rpc_call(abi=lens_abi, contract_address=lens_address, fn_paras=token_address,
                         block_number=block_number,
                         list_call_id=list_call_id, list_rpc_call=list_rpc_call, fn_name="vTokenUnderlyingPrice")
            add_rpc_call(
                abi=comptroller_abi, fn_name='venusSpeeds', contract_address=comptroller_address,
                block_number=block_number, fn_paras=token_address, list_rpc_call=list_rpc_call,
                list_call_id=list_call_id
            )
        return list_rpc_call, list_call_id

    def _decode_apy_lending_pool_function_call(
            self,
            list_rpc_call: list,
            list_call_id: list,
            lens_address: str,
            comptroller_address: str,
            token_addresses: list,
            block_number: int = "latest",
            wrapped_native_token_price: float = 310
    ):
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        ctoken_speeds, borrow_paused_tokens, mint_paused_tokens = {}, {}, {}
        underlying_prices, reserve_tokens_info = {}, []
        for token_address in token_addresses:
            lower_address = token_address.lower()
            metadata_id = f"vTokenMetadata_{lens_address}_{token_address}_{block_number}".lower()
            speeds_call_id = f'venusSpeeds_{comptroller_address}_{token_address}_{block_number}'.lower()
            reserve_tokens_info.append(decoded_data.get(metadata_id))
            underlying_id = f"vTokenUnderlyingPrice_{lens_address}_{token_address}_{block_number}".lower()
            price_token = decoded_data.get(underlying_id)
            underlying_decimals = decoded_data.get(metadata_id)[-1]
            underlying_prices[lower_address] = price_token[1]
            if lower_address in AbnormalVenusPool.decimals.keys():
                underlying_prices[lower_address] /= 10 ** AbnormalVenusPool.decimals.get(lower_address)
            elif underlying_decimals == 8:
                underlying_prices[lower_address] /= 10 ** 20
            elif underlying_decimals == 6:
                underlying_prices[lower_address] /= 10 ** 24
            ctoken_speeds[token_address.lower()] = decoded_data.get(speeds_call_id)
            borrow_paused_tokens[token_address.lower()] = False
            mint_paused_tokens[token_address.lower()] = False

        return {
            "reserve_tokens_info": reserve_tokens_info,
            "ctoken_speeds": ctoken_speeds,
            "borrow_paused_tokens": borrow_paused_tokens,
            "mint_paused_tokens": mint_paused_tokens,
            "underlying_prices": underlying_prices
        }

    def get_apy_lending_pool(
            self,
            lens_address: str,
            comptroller_address: str,
            pool_token_price: float,
            lens_abi: list = VENUS_LENS_ABI,
            comptroller_abi: list = VENUS_COMPTROLLER_ABI,
            pool_decimals: int = 18,
            block_number: int = "latest",
            wrapped_native_token_price: float = 310
    ):
        tokens_interest_rates = dict()
        ctokens = self.get_all_markets(comptroller_address, comptroller_abi, block_number)
        for token in ctokens:
            if token in [ContractAddresses.LUNA.lower(), ContractAddresses.UST.lower(), ContractAddresses.LUNA,
                         ContractAddresses.UST]:
                ctokens.remove(token)
        list_rpc_call, list_call_id = self._encode_apy_lending_pool_function_call(
            lens_address, lens_abi, comptroller_address, comptroller_abi, ctokens, block_number)
        decode_data = self._decode_apy_lending_pool_function_call(
            list_rpc_call, list_call_id, lens_address, comptroller_address, ctokens, block_number)
        mint_paused_tokens = decode_data["mint_paused_tokens"]
        borrow_paused_tokens = decode_data["borrow_paused_tokens"]
        reserve_tokens_info = decode_data["reserve_tokens_info"]
        ctoken_speeds = decode_data["ctoken_speeds"]
        for data in reserve_tokens_info:
            address = data[0].lower()
            underlying_token_price = float(decode_data["underlying_prices"][address]) / 10 ** int(data[13])
            token_info = {
                "token": address,
                "token_decimals": data[12],
                "borrow_rate": data[3],
                "supply_rate": data[2],
                "supply": data[7],
                "borrow": data[5],
                "exchange_rate": data[1],
                "underlying": data[11].lower(),
                "underlying_price": underlying_token_price,
                "underlying_decimals": data[13],
                "speed": ctoken_speeds[address]
            }
            underlying_token = token_info['underlying']
            token_info["mint_paused"] = mint_paused_tokens[address]
            token_info["borrow_paused"] = borrow_paused_tokens[address]
            tokens_interest_rates[underlying_token] = self._calculate_interest_rates(
                token_info, pool_decimals, pool_token_price)

        return tokens_interest_rates

    def get_rewards_balance(
            self,
            wallet_address: str,
            lens_address: str,
            comptroller_address: str,
            pool_token: str = None,
            lens_abi: list = VENUS_LENS_ABI,
            block_number: int = "latest",
    ):
        list_rpc_call = []
        list_call_id = []
        fn_paras = [self.to_checksum(wallet_address), self.to_checksum(comptroller_address)]
        add_rpc_call(abi=lens_abi,
                     fn_paras=fn_paras, block_number=block_number,
                     contract_address=lens_address, fn_name="pendingVenus",
                     list_call_id=list_call_id, list_rpc_call=list_rpc_call)
        get_reward_id = f"pendingVenus_{lens_address}_{fn_paras}_{block_number}".lower()
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        reward = decoded_data[get_reward_id] / 10 ** 18
        return reward

    def get_wallet_deposit_borrow_balance(
            self,
            wallet_address: str,
            lens_address: str,
            reserves_info: dict,
            lens_abi: list = VENUS_LENS_ABI,
            block_number: int = "latest",
            wrapped_native_token_price: float = 310,
            wrapped_native_token: str = None
    ):
        list_rpc_call = []
        list_call_id = []
        for token in reserves_info:
            underlying = token
            value = reserves_info[token]
            if token == WrappedNativeTokens.NATIVE_TOKEN:
                underlying = wrapped_native_token
            add_rpc_call(abi=lens_abi, contract_address=lens_address, fn_paras=value["cToken"],
                         block_number=block_number,
                         list_call_id=list_call_id, list_rpc_call=list_rpc_call, fn_name="vTokenUnderlyingPrice")
            add_rpc_call(abi=CTOKEN_ABI, contract_address=value["cToken"], fn_name="borrowBalanceCurrent",
                         block_number=block_number,
                         fn_paras=wallet_address, list_call_id=list_call_id, list_rpc_call=list_rpc_call)
            add_rpc_call(abi=CTOKEN_ABI, contract_address=value["cToken"], fn_name="balanceOfUnderlying",
                         block_number=block_number,
                         fn_paras=wallet_address, list_call_id=list_call_id, list_rpc_call=list_rpc_call)
            add_rpc_call(abi=ERC20_ABI, contract_address=underlying, fn_name="decimals", block_number=block_number,
                         list_call_id=list_call_id, list_rpc_call=list_rpc_call)

        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        total_borrow, result = 0, {
            "borrow_amount_in_usd": 0,
            "deposit_amount_in_usd": 0,
            "health_factor": 0,
            'reserves_data': {}
        }
        for token in reserves_info:
            underlying = token
            value = reserves_info[token]
            if token == WrappedNativeTokens.NATIVE_TOKEN:
                underlying = wrapped_native_token
            get_total_deposit_id = f"balanceOfUnderlying_{value['cToken']}_{wallet_address}_{block_number}".lower()
            get_total_borrow_id = f"borrowBalanceCurrent_{value['cToken']}_{wallet_address}_{block_number}".lower()
            get_decimals_id = f"decimals_{underlying}_{block_number}".lower()
            decimals = decoded_data[get_decimals_id]
            deposit_amount = decoded_data[get_total_deposit_id] / 10 ** decimals
            borrow_amount = decoded_data[get_total_borrow_id] / 10 ** decimals
            get_underlying_token_price = f"vTokenUnderlyingPrice_{lens_address}_{value['cToken']}_{block_number}".lower()
            token_price = decoded_data.get(get_underlying_token_price)[1] / 10 ** decimals
            if value['cToken'] in AbnormalVenusPool.decimals.keys():
                token_price /= 10 ** AbnormalVenusPool.decimals.get(value['cToken'])

            deposit_amount_in_usd = deposit_amount * token_price
            borrow_amount_in_usd = borrow_amount * token_price
            total_borrow += borrow_amount_in_usd
            result['health_factor'] += deposit_amount_in_usd * value["liquidationThreshold"]
            result['borrow_amount_in_usd'] += borrow_amount_in_usd
            result['deposit_amount_in_usd'] += deposit_amount_in_usd
            if (borrow_amount > 0) or (deposit_amount > 0):
                result['reserves_data'][token] = {
                    "borrow_amount": borrow_amount,
                    "borrow_amount_in_usd": borrow_amount_in_usd,
                    "deposit_amount": deposit_amount,
                    "deposit_amount_in_usd": deposit_amount_in_usd,
                }
        if total_borrow != 0:
            result['health_factor'] /= total_borrow
        else:
            result['health_factor'] = 100
        return result


if __name__ == "__main__":
    import json
    from defi_services.lending_pools.lending_pools_info.bsc.venus_bsc import VENUS_BSC

    service = VenusStateService(provider_uri="https://rpc.ankr.com/bsc")
    reserve_info = service.get_reserves_info(
        lens_address=VENUS_BSC.get("lensAddress"),
        comptroller_address=VENUS_BSC.get("comptrollerAddress"),
        lens_abi=VENUS_LENS_ABI,
        comptroller_abi=VENUS_COMPTROLLER_ABI
    )
    with open("venus_bsc.json", "w") as f:
        f.write(json.dumps(reserve_info, indent=1))



import logging

from query_state_lib.base.mappers.eth_call_mapper import EthCall
from query_state_lib.base.utils.encoder import encode_eth_call_data
from web3 import Web3


def add_rpc_call(abi, fn_name, contract_address, block_number=None, fn_paras=None, list_rpc_call=None,
                 list_call_id=None):
    args = []
    if fn_paras is not None:
        if type(fn_paras) is list:
            args = fn_paras
        else:
            if Web3.isAddress(fn_paras):
                fn_paras = Web3.toChecksumAddress(fn_paras)
            args = [fn_paras]

        call_id = f"{fn_name}_{contract_address}_{fn_paras}_{block_number}".lower()
    else:
        call_id = f"{fn_name}_{contract_address}_{block_number}".lower()

    data_call = encode_eth_call_data(abi=abi, fn_name=fn_name, args=args)
    if block_number:
        eth_call = EthCall(to=Web3.toChecksumAddress(contract_address), block_number=block_number, data=data_call,
                           abi=abi, fn_name=fn_name, id=call_id)
    else:
        eth_call = EthCall(to=Web3.toChecksumAddress(contract_address), data=data_call,
                           abi=abi, fn_name=fn_name, id=call_id)

    if call_id not in list_call_id:
        list_rpc_call.append(eth_call)
        list_call_id.append(call_id)


def decode_data_response(data_responses, list_call_id):
    decoded_datas = {}
    for call_id in list_call_id:
        try:
            response_data = data_responses.get(call_id)
            decoded_data = response_data.decode_result()
        except Exception as e:
            logging.error(f"An exception when decode data from provider: {e}")
            raise

        if len(decoded_data) == 1:
            decoded_datas[call_id] = decoded_data[0]
        else:
            decoded_datas[call_id] = decoded_data
    return decoded_datas



from defi_services.constants.chain_constant import Chain
from defi_services.defi_service import DefiService
from defi_services.lending_pools.lending_pools_info.bsc.venus_bsc import VENUS_BSC
from defi_services.lending_pools.services.venus_state_service import VenusStateService


class VenusService(DefiService):
    name = "venus lending pool"

    def __init__(self, chain_id: str, provider_uri: str):
        super().__init__(chain_id, provider_uri)
        self.state_service = VenusStateService(provider_uri)
        self.defi_constant = self._get_constant()
        self.address = self.defi_constant.get("comptrollerAddress")

    def _get_constant(self):
        if self.chain_id == Chain.bsc:
            return VENUS_BSC
        return None

    def get_pool_token(self):
        return self.defi_constant.get("rewardToken")

    def get_apy_defi_app(self, block_number: int = "latest", **kwargs):
        return self.state_service.get_apy_lending_pool(
            comptroller_address=self.defi_constant.get("comptrollerAddress"),
            lens_address=self.defi_constant.get("lensAddress"),
            pool_token_price=float(kwargs.get("pool_token_price")),
            block_number=block_number
        )

    def get_rewards_balance(self, wallet_address: str, block_number: int = "latest"):
        reward = self.state_service.get_rewards_balance(
            wallet_address=wallet_address,
            comptroller_address=self.defi_constant.get("comptrollerAddress"),
            lens_address=self.defi_constant.get("lensAddress"),
            pool_token=self.defi_constant.get("rewardToken"),
            block_number=block_number
        )
        return {
            self.defi_constant.get("rewardToken"): reward
        }

    def get_wallet_deposit_borrow_balance(self, wallet_address: str, block_number: int = "latest"):
        return self.state_service.get_wallet_deposit_borrow_balance(
            wallet_address=wallet_address,
            lens_address=self.defi_constant.get("lensAddress"),
            reserves_info=self.defi_constant.get("reservesList"),
            block_number=block_number,
            wrapped_native_token=self.get_wrapped_native_token()
        )



from defi_services.constants.chain_constant import Chain
from defi_services.defi_service import DefiService
from defi_services.lending_pools.lending_pools_info.ethereum.aave_v2_eth import AAVE_V2_ETH
from defi_services.lending_pools.lending_pools_info.polygon.aave_v2_polygon import AAVE_V2_POLYGON
from defi_services.lending_pools.services.aave_v2_state_service import AaveV2StateService


class AaveV2Service(DefiService):
    name = "aave v2 lending pool"

    def __init__(self, chain_id: str, provider_uri: str):
        super().__init__(chain_id, provider_uri)
        self.state_service = AaveV2StateService(provider_uri)
        self.defi_constant = self._get_constant()
        self.address = self.defi_constant.get("address")

    def _get_constant(self):
        if self.chain_id == Chain.polygon:
            return AAVE_V2_POLYGON
        elif self.chain_id == Chain.ethereum:
            return AAVE_V2_ETH

        return None

    def get_apy_defi_app(self, block_number: int = "latest", **kwargs):
        return self.state_service.get_apy_lending_pool(
            pool_address=self.defi_constant.get("address"),
            staked_incentive_address=self.defi_constant.get("stakedIncentiveAddress"),
            oracle_address=self.defi_constant.get("oracleAddress"),
            block_number=block_number,
            reserves_info=self.defi_constant.get("reservesList"),
            wrapped_native_token_price=kwargs.get("ethereum_price"),
        )

    def get_rewards_balance(self, wallet_address: str, block_number: int = "latest"):
        reward = self.state_service.get_rewards_balance(
            wallet_address=wallet_address,
            pool_address=self.defi_constant.get("address"),
            staked_incentive_address=self.defi_constant.get("stakedIncentiveAddress"),
            block_number=block_number,
            reserves_info=self.defi_constant.get("reservesList")
        )
        return {
            self.defi_constant.get("rewardToken"): reward
        }

    def get_wallet_deposit_borrow_balance(self, wallet_address: str, block_number: int = "latest", **kwargs):
        return self.state_service.get_wallet_deposit_borrow_balance(
            wallet_address=wallet_address,
            pool_address=self.defi_constant.get("address"),
            oracle_address=self.defi_constant.get("oracleAddress"),
            block_number=block_number,
            reserves_info=self.defi_constant.get("reservesList"),
            wrapped_native_token_price=kwargs.get("ethereum_price"),
        )



from defi_services.constants.chain_constant import Chain
from defi_services.constants.contract_address import ContractAddresses, WrappedNativeTokens
from defi_services.state_service import StateService


class DefiService:
    def __init__(self, chain_id: str, provider_uri: str):
        self.provider_uri = provider_uri
        self.chain_id = chain_id

    def _get_constant(self):
        pass

    def get_wrapped_native_token(self):
        if self.chain_id == Chain.bsc:
            return WrappedNativeTokens.WBNB
        elif self.chain_id == Chain.ethereum:
            return WrappedNativeTokens.WETH
        elif self.chain_id == Chain.polygon:
            return WrappedNativeTokens.WMATIC
        elif self.chain_id == Chain.fantom:
            return WrappedNativeTokens.WFANTOM
        return None

    def get_apy_defi_app(self, block_number: int = "latest"):
        pass

    def get_rewards_balance(self, wallet_address: str, block_number: int = "latest"):
        pass

    def get_wallet_deposit_borrow_balance(self, wallet_address: str, block_number: int = "latest"):
        pass




from defi_services.lending_pools.services.valas_state_service import ValasStateService


class RadiantStateService(ValasStateService):
    def __init__(self, provider_uri: str):
        super().__init__(provider_uri)


if __name__ == "__main__":
    import json
    from defi_services.lending_pools.lending_pools_info.bsc.radiant_bsc import RADIANT_BSC
    from defi_services.lending_pools.lending_pools_info.arbitrum.radiant_arbitrum import RADIANT_ARB
    from defi_services.abis.lending_pool.lending_pool_abi import LENDING_POOL_ABI

    service = RadiantStateService(provider_uri="https://rpc.ankr.com/bsc")
    reserve_info = service.get_reserves_info(RADIANT_BSC.get("address"), LENDING_POOL_ABI)
    with open("radiant.json", "w") as f:
        f.write(json.dumps(reserve_info, indent=1))
    service = RadiantStateService(provider_uri="https://rpc.ankr.com/arbitrum")
    reserve_info = service.get_reserves_info(RADIANT_ARB.get("address"), LENDING_POOL_ABI)
    with open("radiant.json", "w") as f:
        f.write(json.dumps(reserve_info, indent=1))



from defi_services.lending_pools.services.cream_state_service import CreamStateService
from defi_services.utils.batch_queries_service import add_rpc_call, decode_data_response
from defi_services.abis.lending_pool.cream_lens_abi import CREAM_LENS_ABI
from defi_services.abis.lending_pool.cream_comptroller_abi import CREAM_COMPTROLLER_ABI


class CompoundStateService(CreamStateService):
    def __init__(self, provider_uri: str):
        super().__init__(provider_uri)

    def get_rewards_balance(
            self,
            wallet_address: str,
            lens_address: str,
            comptroller_address: str,
            pool_token: str,
            lens_abi: list = CREAM_LENS_ABI,
            block_number: int = "latest",
    ):
        list_rpc_call = []
        list_call_id = []
        fn_paras = [self.to_checksum(pool_token),
                    self.to_checksum(comptroller_address),
                    self.to_checksum(wallet_address)]
        add_rpc_call(abi=lens_abi,
                     fn_paras=fn_paras,
                     contract_address=lens_address,
                     fn_name="getCompBalanceMetadataExt", block_number=block_number,
                     list_call_id=list_call_id, list_rpc_call=list_rpc_call)
        get_reward_id = f"getCompBalanceMetadataExt_{lens_address}_{fn_paras}_{block_number}".lower()
        data_response = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=100)
        decoded_data = decode_data_response(data_response, list_call_id)
        reward = decoded_data[get_reward_id][-1] / 10 ** 18
        return reward

    def get_apy_lending_pool(
            self,
            lens_address: str,
            comptroller_address: str,
            pool_token_price: float,
            lens_abi: list = CREAM_LENS_ABI,
            comptroller_abi: list = CREAM_COMPTROLLER_ABI,
            pool_decimals: int = 18,
            block_number: int = "latest",
            wrapped_native_token_price: float = 310
    ):
        tokens_interest_rates = dict()
        ctokens = self.get_all_markets(comptroller_address, comptroller_abi, block_number)
        list_rpc_call, list_call_id = self._encode_apy_lending_pool_function_call(
            lens_address, lens_abi, comptroller_address, comptroller_abi, ctokens, block_number)
        decode_data = self._decode_apy_lending_pool_function_call(
            list_rpc_call, list_call_id, lens_address, comptroller_address, ctokens, block_number,
            wrapped_native_token_price)
        mint_paused_tokens = decode_data["mint_paused_tokens"]
        borrow_paused_tokens = decode_data["borrow_paused_tokens"]
        reserve_tokens_info = decode_data["reserve_tokens_info"]
        ctoken_speeds = decode_data["ctoken_speeds"]
        for data in reserve_tokens_info:
            address = data[0].lower()
            underlying_token_price = float(decode_data["underlying_prices"][address]) / 10 ** int(data[13])
            token_info = {
                "token": address,
                "token_decimals": data[12],
                "borrow_rate": data[3],
                "supply_rate": data[2],
                "supply": data[7],
                "borrow": data[5],
                "exchange_rate": data[1],
                "underlying": data[11].lower(),
                "underlying_price": underlying_token_price,
                "underlying_decimals": data[13],
                "speed": ctoken_speeds[address]
            }
            underlying_token = token_info['underlying']
            token_info["mint_paused"] = mint_paused_tokens[address]
            token_info["borrow_paused"] = borrow_paused_tokens[address]
            tokens_interest_rates[underlying_token] = self._calculate_interest_rates(
                token_info, pool_decimals, pool_token_price)

        return tokens_interest_rates


if __name__ == "__main__":
    import json
    from defi_services.lending_pools.lending_pools_info.ethereum.compound_eth import COMPOUND_ETH

    service = CompoundStateService(provider_uri="https://rpc.ankr.com/eth")
    reserve_info = service.get_reserves_info(
        lens_address=COMPOUND_ETH.get("lensAddress"),
        comptroller_address=COMPOUND_ETH.get("comptrollerAddress"),
        lens_abi=CREAM_LENS_ABI,
        comptroller_abi=CREAM_COMPTROLLER_ABI
    )
    with open("compound_bsc.json", "w") as f:
        f.write(json.dumps(reserve_info, indent=1))



VALUATOR_ABI = [
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_token",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "_amount",
                "type": "uint256"
            }
        ],
        "name": "valuation",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    }
]



from defi_services.lending_pools.aave_v2_service import AaveV2Service
from defi_services.lending_pools.aave_v3_service import AaveV3Service
from defi_services.lending_pools.compound_service import CompoundService
from defi_services.lending_pools.cream_service import CreamService
from defi_services.lending_pools.geist_service import GeistService
from defi_services.lending_pools.radiant_service import RadiantService
from defi_services.lending_pools.trava_service import TravaService
from defi_services.lending_pools.valas_service import ValasService
from defi_services.lending_pools.venus_service import VenusService

LENDING_POOL_SERVICE = {
    "0x38": [TravaService, ValasService, VenusService, CreamService, RadiantService, AaveV3Service],
    "0xfa": [TravaService, GeistService, AaveV3Service],
    "0x1": [TravaService, AaveV2Service, CompoundService, AaveV3Service],
    "0x89": [AaveV2Service, AaveV3Service],
    "0xa4b1": [AaveV3Service, RadiantService],
    "0xa": [AaveV3Service]
}



import re

import setuptools

with open("src/defi_services/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="defi-services-lib",
    version=version,
    author="Viet-Bang Pham",
    author_email="phamvietbang2965@gmail.com",
    description="Calculate apy, apr, and wallet information,... in decentralized applications.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phamvietbang/defi-services-lib",
    project_urls={"Bug Tracker": "https://github.com/phamvietbang/defi-services-lib", },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # web3.py doesn't work on 3.5.2 and less (https://github.com/ethereum/web3.py/issues/1012)
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires='>=3.6',
    install_requires=[
        "web3==5.31.1",
        "QueryStateLib==1.1.4"
    ],
)



from defi_services.lending_pools.services.valas_state_service import ValasStateService


class GeistStateService(ValasStateService):
    def __init__(self, provider_uri: str):
        super().__init__(provider_uri)


if __name__ == "__main__":
    import json
    from defi_services.lending_pools.lending_pools_info.fantom.geist_ftm import GEIST_ETH
    from defi_services.abis.lending_pool.lending_pool_abi import LENDING_POOL_ABI

    service = GeistStateService(provider_uri="https://rpc.ankr.com/fantom")
    reserve_info = service.get_reserves_info(GEIST_ETH.get("address"), LENDING_POOL_ABI)
    with open("geist.json", "w") as f:
        f.write(json.dumps(reserve_info, indent=1))
